#!/usr/bin/env python

"""
Inspect source files, obtaining details of classes and attributes.

Copyright (C) 2007, 2008 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.

--------

The results of inspecting a module are as follows:

Constants
---------

All constants identified within the code shall be registered.

Classes
-------

All global classes shall be registered; local classes (within functions) or
nested classes (within classes) are not currently registered.

Base classes must be detected and constant.

All classes without bases are made to inherit from __builtins__.object in order
to support some standard methods.

Functions
---------

All functions and lambda definitions shall be registered.

Namespaces
----------

Modules define their own "global" namespace, within which classes, functions
and lambda definitions establish a hierarchy of namespaces.

Only local, global and built-in namespaces are recognised; closures are not
supported.

Assignments
-----------

Name assignment and attribute assignment involving modules and classes cause
names to be associated with values within namespaces.

Any assignments within loops are considered to cause the targets of such
assignments to provide non-constant values.

Assignments to names are only really considered to cause the targets of such
assignments to provide constant values if the targets reside in the module
namespace or in class namespaces, subject to the above conditions.

Assignments to names within functions are not generally considered to cause the
targets of such assignments to provide constant values since functions can be
invoked many times with different inputs. However, there may be benefits in
considering a local to be constant within a single invocation.
"""

from micropython.common import *
from micropython.data import *
import compiler.ast
from compiler.visitor import ASTVisitor

# Program visitors.

class InspectedModule(ASTVisitor, Module):

    """
    An inspected module, providing core details via the Module superclass, but
    capable of being used as an AST visitor.
    """

    def __init__(self, name, importer):

        """
        Initialise this visitor with a module 'name' and an 'importer' which is
        used to provide access to other modules when required.
        """

        ASTVisitor.__init__(self)
        Module.__init__(self, name)
        self.visitor = self

        # Import machinery links.

        self.importer = importer
        self.builtins = self.importer.modules.get("__builtins__")
        self.loaded = 0

        # Constant records.

        self.constant_values = importer.constant_values

        # Current expression state.

        self.expr = None

        # Namespace state.

        self.in_init = 0        # Find instance attributes in __init__ methods.
        self.in_method = 0      # Find instance attributes in all methods.
        self.in_loop = 0        # Note loop "membership", affecting assignments.
        self.namespaces = []
        self.module = None

    def parse(self, filename):

        "Parse the file having the given 'filename'."

        module = compiler.parseFile(filename)
        self.process(module)

    def process(self, module):

        "Process the given 'module'."

        self.astnode = self.module = module
        processed = self.dispatch(module)
        if self.has_key("__all__"):
            all = self["__all__"]
            if isinstance(all, compiler.ast.List):
                for n in all.nodes:
                    self[n.value] = self.importer.add_module(self.name + "." + n.value)
        return processed

    def vacuum(self):

        "Vacuum the module namespace, removing unloaded module references."

        for name, value in self.items():
            if isinstance(value, Module) and not value.loaded:
                del self[name]

            # Complain about globals not initialised at the module level.

            if isinstance(value, Global):
                print "Warning: global %r in module %r not initialised at the module level." % (name, self.name)

    # Namespace methods.

    def store(self, name, obj):

        "Record attribute or local 'name', storing 'obj'."

        if not self.namespaces:
            self.set(name, obj, not self.in_loop)
        else:
            self.namespaces[-1].set(name, obj, not self.in_loop)

        # Record all non-local objects.

        if not (self.namespaces and isinstance(self.namespaces[-1], Function)):
            self.all_objects.add(obj)

    def store_lambda(self, obj):

        "Store a lambda function 'obj'."

        self.all_objects.add(obj)

    def store_module_attr(self, name, module):

        """
        Record module attribute 'name' in the given 'module' using the current
        expression.
        """

        if isinstance(self.expr, Attr):
            assigned_value = self.expr.value
        else:
            assigned_value = self.expr

        module.set(name, assigned_value, 0)

    def store_class_attr(self, name):

        """
        Record class attribute 'name' in the current class using the current
        expression.
        """

        if self.in_method and self.namespaces[-2].has_key(name):

            if isinstance(self.expr, Attr):
                assigned_value = self.expr.value
            else:
                assigned_value = self.expr

            self.namespaces[-2].set(name, assigned_value, 0)
            return 1

        return 0

    def store_instance_attr(self, name):

        "Record instance attribute 'name' in the current class."

        if self.in_method:

            # Current namespace is the function.
            # Previous namespace is the class.

            self.namespaces[-2].add_instance_attribute(name)

    def get_parent(self):

        "Return the parent (or most recent) namespace currently exposed."

        return (self.namespaces[-1:] or [self])[0]

    # Visitor methods.

    def default(self, node, *args):
        raise InspectError(self.full_name(), node, "Node class %r is not supported." % node.__class__)

    def dispatch(self, node, *args):
        return ASTVisitor.dispatch(self, node, *args)

    def NOP(self, node):
        for n in node.getChildNodes():
            self.dispatch(n)
        return None

    def OP(self, node):
        for n in node.getChildNodes():
            self.dispatch(n)
        return Instance()

    def _make_constant(self, value):
        if not self.constant_values.has_key(value):
            const = Const(value)
            self.constant_values[value] = const
        return self.constant_values[value]

    def _visitFunction(self, node, name):

        """
        Return a function object for the function defined by 'node' with the
        given 'name'. If a lambda expression is being visited, 'name' should be
        None.
        """

        function = Function(
            name,
            self.get_parent(),
            node.argnames,
            node.defaults,
            (node.flags & 4 != 0),
            (node.flags & 8 != 0),
            self,
            node
            )

        # Make a back reference from the node for code generation.

        node.unit = function

        # Process the defaults.

        for n in node.defaults:
            self.expr = self.dispatch(n)
            if isinstance(self.expr, Attr):
                function.store_default(self.expr.value)
            else:
                function.store_default(self.expr)

        # Enter the function.

        self.namespaces.append(function)

        # Current namespace is the function.
        # Previous namespace is the class.

        if len(self.namespaces) > 1 and isinstance(self.namespaces[-2], Class):
            if name == "__init__":
                self.in_init = 1
            self.in_method = 1

        self.dispatch(node.code)
        self.in_init = 0
        self.in_method = 0
        self.namespaces.pop()

        if name is not None:
            self.store(name, function)
        else:
            self.store_lambda(function)
            if node.defaults:
                return None # NOTE: See lambda code in the ast module.

        return function

    visitAdd = OP

    visitAnd = OP

    visitAssert = NOP

    def visitAssign(self, node):
        self.expr = self.dispatch(node.expr)
        for n in node.nodes:
            self.dispatch(n)
        return None

    def visitAssAttr(self, node):
        expr = self.dispatch(node.expr)
        if isinstance(expr, Attr):
            if expr.name == "self":
                if not self.store_class_attr(node.attrname):
                    self.store_instance_attr(node.attrname)
            elif isinstance(expr.value, Module):
                self.store_module_attr(node.attrname, expr.value)
                print "Warning: attribute %r of module %r set outside the module." % (node.attrname, expr.value.name)
        return None

    def visitAssList(self, node):
        for i, n in enumerate(node.nodes):
            self.dispatch(n)
            self._make_constant(i) # for __getitem__(i) at run-time
        return None

    def visitAssName(self, node):
        self.store(node.name, self.expr)
        return None

    visitAssTuple = visitAssList

    visitAugAssign = OP

    visitBackquote = OP

    visitBitand = OP

    visitBitor = OP

    visitBitxor = OP

    visitBreak = NOP

    visitCallFunc = OP

    def visitClass(self, node):

        """
        Register the class at the given 'node' subject to the restrictions
        mentioned in the module docstring.
        """

        if self.namespaces:
            print "Class %r in %r is not global: ignored." % (node.name, self.namespaces[-1].full_name())
            return None
        else:
            cls = Class(node.name, self.get_parent(), self, node)

            # Visit the base class expressions, attempting to find concrete
            # definitions of classes.

            for base in node.bases:
                expr = self.dispatch(base)
                if isinstance(expr, Attr):
                    if expr.assignments != 1:
                        raise InspectError(self.full_name(), node,
                            "Base class %r for %r is not constant." % (base, cls.full_name()))
                    else:
                        cls.add_base(expr.value)
                else: # if expr is None:
                    raise InspectError(self.full_name(), node,
                        "Base class %r for %r is not found: it may be hidden in some way." % (base, cls.full_name()))

            # NOTE: Potentially dubious measure to permit __init__ availability.
            # If no bases exist, adopt the 'object' class.

            if not node.bases and not (self.name == "__builtins__" and node.name == "object") :
                expr = self.dispatch(compiler.ast.Name("object"))
                cls.add_base(expr.value)

            # Make a back reference from the node for code generation.

            node.unit = cls

            # Make an entry for the class.

            self.store(node.name, cls)

            # Process the class body.

            self.namespaces.append(cls)
            self.dispatch(node.code)
            self.namespaces.pop()

            return cls

    visitCompare = OP

    def visitConst(self, node):
        return self._make_constant(node.value)

    visitContinue = NOP

    visitDecorators = NOP

    visitDict = OP

    visitDiscard = NOP

    visitDiv = OP

    visitEllipsis = NOP

    visitExec = NOP

    visitExpression = OP

    visitFloorDiv = OP

    def visitFor(self, node):
        self.in_loop = 1
        self.NOP(node)
        self.in_loop = 0
        return None

    def visitFrom(self, node):
        module = self.importer.load(node.modname, 1)

        #if module is None:
        #    print "Warning:", node.modname, "not imported."

        for name, alias in node.names:
            if name != "*":
                if module is not None and module.namespace.has_key(name):
                    attr = module[name]
                    self.store(alias or name, attr.value)
                    if isinstance(attr, Module) and not attr.loaded:
                        self.importer.load(attr.name)

                # Support the import of names from missing modules.

                else:
                    self.store(alias or name, UnresolvedName(name, node.modname, self))
            else:
                if module is not None:
                    for n in module.namespace.keys():
                        attr = module[n]
                        self.store(n, attr.value)
                        if isinstance(attr, Module) and not attr.loaded:
                            self.importer.load(attr.name)

        return None

    def visitFunction(self, node):
        return self._visitFunction(node, node.name)

    visitGenExpr = OP

    visitGenExprFor = NOP

    visitGenExprIf = NOP

    visitGenExprInner = NOP

    def visitGetattr(self, node):
        expr = self.dispatch(node.expr)
        if isinstance(expr, Attr):
            value = expr.value
            if isinstance(value, (Class, Module)):
                return value.namespace.get(node.attrname)
            elif isinstance(value, UnresolvedName):
                return UnresolvedName(node.attrname, value.full_name(), self)
            else:
                return None
        if self.builtins is not None:
            return self.builtins.get(node.attrname)
        else:
            return UnresolvedName(node.attrname, value.full_name(), self)

    def visitGlobal(self, node):
        if self.namespaces:
            for name in node.names:
                self.namespaces[-1].make_global(name)

                # Record a global entry for the name in the module.

                if not self.has_key(name):
                    self[name] = Global()

    def visitIf(self, node):
        for test, body in node.tests:
            self.dispatch(body)
        if node.else_ is not None:
            self.dispatch(node.else_)
        return None

    visitIfExp = NOP

    def visitImport(self, node):
        for name, alias in node.names:
            if alias is not None:
                self.store(alias, self.importer.load(name, 1) or UnresolvedName(None, name, self))
            else:
                self.store(name.split(".")[0], self.importer.load(name) or UnresolvedName(None, name.split(".")[0], self))

        return None

    visitInvert = OP

    def visitKeyword(self, node):
        self.dispatch(node.expr)
        self._make_constant(node.name)
        self.keyword_names.add(node.name)
        return None

    def visitLambda(self, node):
        return self._visitFunction(node, None)

    visitLeftShift = OP

    visitList = OP

    visitListComp = OP

    visitListCompFor = NOP

    visitListCompIf = NOP

    visitMod = OP

    def visitModule(self, node):
        return self.dispatch(node.node)

    visitMul = OP

    def visitName(self, node):
        name = node.name
        if name == "None":
            return self._make_constant(None)
        elif self.namespaces and self.namespaces[-1].has_key(name):
            return self.namespaces[-1][name]
        elif self.has_key(name):
            return self[name]
        elif self.builtins is not None and self.builtins.has_key(name):
            return self.builtins[name]
        else:
            return None

    visitNot = OP

    visitOr = OP

    visitPass = NOP

    visitPower = OP

    visitPrint = NOP

    visitPrintnl = NOP

    visitRaise = NOP

    visitReturn = NOP

    visitRightShift = OP

    visitSlice = OP

    visitSliceobj = OP

    def visitStmt(self, node):
        for n in node.nodes:
            self.dispatch(n)
        return None

    visitSub = OP

    visitSubscript = OP

    def visitTryExcept(self, node):
        self.dispatch(node.body)
        for name, var, n in node.handlers:

            # Establish the local for the handler.

            if var is not None:
                self.dispatch(var)
            if n is not None:
                self.dispatch(n)
        if node.else_ is not None:
            self.dispatch(node.else_)
        return None

    visitTryFinally = NOP

    visitTuple = OP

    visitUnaryAdd = OP

    visitUnarySub = OP

    def visitWhile(self, node):
        self.in_loop = 1
        self.NOP(node)
        self.in_loop = 0
        return None

    visitWith = NOP

    visitYield = NOP

class Global:

    """
    A reference to an object assigned to a global from outside the module
    top-level.
    """

    pass

# vim: tabstop=4 expandtab shiftwidth=4
